//
//  AppDelegate.swift
//  TextLCD
//
//  Created by Phidgets on 2018-02-16.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate, NSTextFieldDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = LCD()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    //GUI
    @IBOutlet weak var contrastSlider: NSSlider!
    @IBOutlet weak var minContrastLabel: NSTextField!
    @IBOutlet weak var maxContrastLabel: NSTextField!
    @IBOutlet weak var contrastLabel: NSTextField!
    @IBOutlet weak var backlightSlider: NSSlider!
    @IBOutlet weak var minBacklightLabel: NSTextField!
    @IBOutlet weak var maxBacklightLabel: NSTextField!
    @IBOutlet weak var backlightLabel: NSTextField!
    @IBOutlet weak var screenSizePopup: NSPopUpButton!
    @IBOutlet weak var initButton: NSButton!
    @IBOutlet weak var settingsBox: NSBox!
    @IBOutlet weak var textView: NSView!
    
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0
    
    //bitmaps
    let heart:[UInt8] = [0,0,0,0,0,
                         0,1,0,1,0,
                         1,1,1,1,1,
                         1,1,1,1,1,
                         0,1,1,1,0,
                         0,0,1,0,0,
                         0,0,0,0,0,
                         0,0,0,0,0]
    
    let smiley:[UInt8] =  [0,0,0,0,0,
                           0,1,0,1,0,
                           0,0,0,0,0,
                           0,0,1,0,0,
                           0,0,0,0,0,
                           1,0,0,0,1,
                           0,1,1,1,0,
                           0,0,0,0,0]

    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! LCD
            let deviceID = try attachedDevice.getDeviceID()
            let minContrast = try attachedDevice.getMinContrast()
            let maxContrast = try attachedDevice.getMaxContrast()
            let contrast = try attachedDevice.getContrast()
            let minBacklight = try attachedDevice.getMinBacklight()
            let maxBacklight = try attachedDevice.getMaxBacklight()
            let backlight = try attachedDevice.getBacklight()
            
            if(deviceID == DeviceID.PN_1204){
                try ch.setScreenSize(LCDScreenSize.noScreen)
            }
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: attachedDevice)
                self.settingsBox.isHidden = false
                
                self.contrastSlider.minValue = minContrast
                self.contrastSlider.maxValue = maxContrast
                self.contrastSlider.doubleValue = contrast
                self.minContrastLabel.stringValue = String(format:"%.2f",minContrast)
                self.maxContrastLabel.stringValue = String(format:"%.2f",maxContrast)
                self.contrastLabel.stringValue = String(format:"%.2f",contrast)
                
                self.backlightSlider.minValue = minBacklight
                self.backlightSlider.maxValue = maxBacklight
                self.backlightSlider.doubleValue = backlight
                self.minBacklightLabel.stringValue = String(format:"%.2f",minBacklight)
                self.maxBacklightLabel.stringValue = String(format:"%.2f",maxBacklight)
                self.backlightLabel.stringValue = String(format:"%.2f",backlight)
                
                if(deviceID == DeviceID.PN_1204){
                    self.screenSizePopup.isHidden = false
                    self.initButton.isHidden = false
                    self.screenSizePopup.addItem(withTitle: "No Screen")
                    self.screenSizePopup.addItem(withTitle: "1x8")
                    self.screenSizePopup.addItem(withTitle: "2x8")
                    self.screenSizePopup.addItem(withTitle: "1x16")
                    self.screenSizePopup.addItem(withTitle: "2x16")
                    self.screenSizePopup.addItem(withTitle: "4x16")
                    self.screenSizePopup.addItem(withTitle: "2x20")
                    self.screenSizePopup.addItem(withTitle: "4x20")
                    self.screenSizePopup.addItem(withTitle: "2x24")
                    self.screenSizePopup.addItem(withTitle: "1x40")
                    self.screenSizePopup.addItem(withTitle: "2x40")
                    self.screenSizePopup.addItem(withTitle: "4x40")
                    self.screenSizePopup.selectItem(at: 0)
                }
                else{
                    self.screenSizePopup.isHidden = true
                    self.initButton.isHidden = true
                }
                
                self.updateDisplay()
                
                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.settingsBox.isHidden = true
            
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    //MARK: GUI Controls
    @IBAction func setContrast(_ sender: Any) {
        let slider = sender as! NSSlider
        do{
            try ch.setContrast(slider.doubleValue)
            contrastLabel.stringValue = String(format:"%.2f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setBacklight(_ sender: Any) {
        let slider = sender as! NSSlider
        do{
            try ch.setBacklight(slider.doubleValue)
            backlightLabel.stringValue = String(format:"%.2f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setScreenSize(_ sender: Any) {
        do{
            let popup = sender as! NSPopUpButton
            switch(popup.indexOfSelectedItem){
            case 0:
                try ch.setScreenSize(LCDScreenSize.noScreen)
                break
            case 1:
                try ch.setScreenSize(LCDScreenSize.dimensions_1x8)
                break
            case 2:
                try ch.setScreenSize(LCDScreenSize.dimensions_2x8)
                break
            case 3:
                try ch.setScreenSize(LCDScreenSize.dimensions_1x16)
                break
            case 4:
                try ch.setScreenSize(LCDScreenSize.dimensions_2x16)
                break
            case 5:
                try ch.setScreenSize(LCDScreenSize.dimensions_4x16)
                break
            case 6:
                try ch.setScreenSize(LCDScreenSize.dimensions_2x20)
                break
            case 7:
                try ch.setScreenSize(LCDScreenSize.dimensions_4x20)
                break
            case 8:
                try ch.setScreenSize(LCDScreenSize.dimensions_2x24)
                break
            case 9:
                try ch.setScreenSize(LCDScreenSize.dimensions_1x40)
                break
            case 10:
                try ch.setScreenSize(LCDScreenSize.dimensions_2x40)
                break
            case 11:
                try ch.setScreenSize(LCDScreenSize.dimensions_4x40)
                break
            default:
                break
            }
            updateDisplay()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setCursor(_ sender: Any) {
        do{
            let check = sender as! NSButton
            try ch.setCursorOn(check.state == NSControl.StateValue.on ? true : false)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setCursorBlink(_ sender: Any) {
        do{
            let check = sender as! NSButton
            try ch.setCursorBlink(check.state == NSControl.StateValue.on ? true : false)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setCustomCharacters(_ sender: Any) {
        do{
            try ch.setCharacterBitmap(font: LCDFont.dimensions_5x8, character: "\u{8}", bitmap: heart)
            try ch.setCharacterBitmap(font: LCDFont.dimensions_5x8, character: "\u{9}", bitmap: smiley)
            try ch.writeText(font: LCDFont.dimensions_5x8, xPosition: 0, yPosition: 0, text: "\u{8} Custom chars! \u{9}")
            try ch.flush()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func initLCD(_ sender: Any) {
        do{
            try ch.initialize()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    
    @IBAction func clear(_ sender: Any) {
        do{
            try ch.clear()
            try ch.flush()
            
            for view in self.textView.subviews{
                let textField = view as! NSTextField
                textField.stringValue = ""
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
    
    //MARK: Misc
    func updateDisplay(){
        do{
            let height = try ch.getHeight()
            DispatchQueue.main.async {
                for view in self.textView.subviews{
                    let textField = view as! NSTextField
                    textField.isEnabled = false
                    textField.backgroundColor = #colorLiteral(red: 0.9125197593, green: 0.9125197593, blue: 0.9125197593, alpha: 1)
                    if(textField.tag < height){
                        textField.isEnabled = true
                        textField.backgroundColor = NSColor.clear
                    }
                }
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Textfield
    override func controlTextDidChange(_ obj: Notification) {
        let textField:NSTextField = obj.object as! NSTextField
        do{
            try ch.clear()
            for view in textView.subviews{
                let field = view as! NSTextField
                if(field.isEnabled){
                    try ch.writeText(font: LCDFont.dimensions_5x8, xPosition: 0, yPosition: field.tag, text: field.stringValue)
                }
            }
            //write here again so cursor position is correct
            try ch.writeText(font: LCDFont.dimensions_5x8, xPosition: 0, yPosition: textField.tag, text: textField.stringValue)
            try ch.flush()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
}

